// HtmlExplorerBar.cpp : Implementation of CHtmlBar

#include "stdafx.h"

#include "HtmlBar.h"
#include "HtmlExplorerBar.h"

#include "HtmlFormat.h"
#include "ScriptFormat.h"
#include "JScriptFormat.h"
#include "VBScriptFormat.h"
#include "CssFormat.h"
#include "PlainTextFormat.h"


// For the little RTF stream hack.
CString CHtmlBar::m_sStream;
long    CHtmlBar::m_lStreamIndex = 0;


/////////////////////////////////////////////////////////////////////////////
// CHtmlBar

HRESULT CHtmlBar::FinalConstruct()
{
   m_hWnd = NULL;
   m_hwndParent = NULL;
   m_bFocus = FALSE;
   m_dwViewMode = 0;
   m_dwBandID = 0;
   m_pFindDlg = NULL;
   m_pRichEditOle = NULL;
   return S_OK;
}

void CHtmlBar::FinalRelease()
{
   if( m_pRichEditOle != NULL ) m_pRichEditOle->Release();
}


/////////////////////////////////////////////////////////////////////////////
// IOleWindow methods

STDMETHODIMP CHtmlBar::GetWindow(HWND* phWnd)
{
   ATLASSERT(phWnd);
   *phWnd = m_hWnd;
   return S_OK;
}

STDMETHODIMP CHtmlBar::ContextSensitiveHelp(BOOL)
{
   return E_NOTIMPL;
}


/////////////////////////////////////////////////////////////////////////////
// IDockingWindow methods

STDMETHODIMP CHtmlBar::ShowDW(BOOL fShow)
{
   if(m_hWnd) {
      if(fShow) {
         // Show our window
         ::ShowWindow(m_hWnd, SW_SHOW);
         InitTree();
      }
      else {
         // Hide our window
         RemoveTree();
         ::ShowWindow(m_hWnd, SW_HIDE);
      }
   }
   return S_OK;
}

STDMETHODIMP CHtmlBar::CloseDW(DWORD dwReserved)
{
   ShowDW(FALSE);
   if( ::IsWindow(m_hWnd) ) ::DestroyWindow(m_hWnd);
   m_hWnd = NULL;  
   return S_OK;
}

STDMETHODIMP CHtmlBar::ResizeBorderDW(LPCRECT prcBorder, IUnknown* punkToolbarSite, BOOL fReserved)
{
   // This method is never called for Band Objects.
   return E_NOTIMPL;
}


/////////////////////////////////////////////////////////////////////////////
// IDeskBand methods
STDMETHODIMP CHtmlBar::GetBandInfo(DWORD dwBandID, DWORD dwViewMode, DESKBANDINFO* pdbi)
{
   if( pdbi!=NULL ) {
      m_dwBandID = dwBandID;
      m_dwViewMode = dwViewMode;
      if(pdbi->dwMask & DBIM_MINSIZE) {
         if(DBIF_VIEWMODE_FLOATING & dwViewMode) {
            pdbi->ptMinSize.x = 200;
            pdbi->ptMinSize.y = 400;
         }
         else {
            pdbi->ptMinSize.x = MIN_SIZE_X;
            pdbi->ptMinSize.y = MIN_SIZE_Y;
         }
      }
      if(pdbi->dwMask & DBIM_MAXSIZE) {
         pdbi->ptMaxSize.x = -1;
         pdbi->ptMaxSize.y = -1;
      }
      if(pdbi->dwMask & DBIM_INTEGRAL) {
         pdbi->ptIntegral.x = 1;
         pdbi->ptIntegral.y = 1;
      }
      if(pdbi->dwMask & DBIM_ACTUAL) {
         pdbi->ptActual.x = 100;
         pdbi->ptActual.y = 100;
      }
      if(pdbi->dwMask & DBIM_TITLE) {
         TCHAR szTitle[100];
         ::LoadString(_Module.GetResourceInstance(), IDS_TITLE, szTitle, 100);
         USES_CONVERSION;
         lstrcpyW(pdbi->wszTitle, T2CW(szTitle));
      }
      if(pdbi->dwMask & DBIM_MODEFLAGS) {
         pdbi->dwModeFlags = DBIMF_NORMAL;
         pdbi->dwModeFlags |= DBIMF_VARIABLEHEIGHT;
      }
      if(pdbi->dwMask & DBIM_BKCOLOR) {
         // Use the default background color by removing this flag.
         pdbi->dwMask &= ~DBIM_BKCOLOR;
      }
      return S_OK;
   }

   return E_INVALIDARG;
}


/////////////////////////////////////////////////////////////////////////////
// IInputObject methods

STDMETHODIMP CHtmlBar::UIActivateIO(BOOL fActivate, LPMSG pMsg)
{
   if(fActivate) ::SetFocus(m_hWnd);
   return S_OK;
}

STDMETHODIMP CHtmlBar::HasFocusIO(void)
{
   if(m_bFocus) return S_OK; else return S_FALSE;
}

STDMETHODIMP CHtmlBar::TranslateAcceleratorIO(LPMSG)
{
   return S_FALSE;
}


/////////////////////////////////////////////////////////////////////////////
// IObjectWithSite methods

STDMETHODIMP CHtmlBar::SetSite(IUnknown* punkSite)
{
   if( punkSite!=NULL ) {
   
      m_hwndParent = NULL;
      CComQIPtr<IOleWindow> spOleWindow = punkSite;
      if( spOleWindow==NULL ) return E_UNEXPECTED;

      spOleWindow->GetWindow(&m_hwndParent);
      if( m_hwndParent==NULL ) return E_FAIL;
      if( !RegisterAndCreateWindow() ) return E_FAIL;

      CComQIPtr<IServiceProvider> spSP(punkSite);
      if( spSP==NULL ) return E_FAIL;

      // Get to the Web Browser app
      HRESULT Hr;
      Hr = spSP->QueryService(IID_IWebBrowserApp, &m_spWebBrowser2);
      if( FAILED(Hr) ) return E_FAIL;
      
      // Retrieve and store the IConnectionPointerContainer pointer
      m_spCPC = m_spWebBrowser2;
      if (m_spCPC == NULL) return E_FAIL;
      // Sign up for browser events
      // TODO: No real need for these before the actual Explorer Bar
      //       is shown. Consider placing this in the ShowDW() method.
      BrowserConnect();
   }
   else {
      // No more browser events, please.
      BrowserDisconnect();

      if( m_spWebBrowser2!=NULL ) m_spWebBrowser2.Release();
      if( m_spCPC!=NULL ) m_spCPC.Release();
   }

   // Get and keep the IInputObjectSite pointer.
   m_spSite = punkSite;

   return S_OK;
}

STDMETHODIMP CHtmlBar::GetSite(REFIID riid, LPVOID *ppvReturn)
{
   ATLASSERT(ppvReturn);
   if( m_spSite ) return m_spSite->QueryInterface(riid, ppvReturn);
   return E_FAIL;
}


/////////////////////////////////////////////////////////////////////////////
// IPersistStream methods

STDMETHODIMP CHtmlBar::GetClassID(LPCLSID pClassID)
{
   ATLASSERT(pClassID);
   *pClassID = CLSID_HtmlBar;
   return S_OK;
}

STDMETHODIMP CHtmlBar::IsDirty(void)
{
   return S_FALSE;
}

STDMETHODIMP CHtmlBar::Load(LPSTREAM)
{
   return S_OK;
}

STDMETHODIMP CHtmlBar::Save(LPSTREAM, BOOL)
{
   return S_OK;
}

STDMETHODIMP CHtmlBar::GetSizeMax(ULARGE_INTEGER*)
{
   return E_NOTIMPL;
}


/////////////////////////////////////////////////////////////////////////////
// IContextMenu methods

STDMETHODIMP CHtmlBar::QueryContextMenu(HMENU hMenu,
                                        UINT indexMenu,
                                        UINT idCmdFirst,
                                        UINT idCmdLast,
                                        UINT uFlags)

{
   if( (CMF_DEFAULTONLY & uFlags)==0 ) {
      CString s;
      s.LoadString(IDS_COPY);
      ::InsertMenu( hMenu, 
         indexMenu++, 
         MF_STRING | MF_BYPOSITION, 
         idCmdFirst + IDM_CMD_COPY, 
         s);
      s.LoadString(IDS_PRINT);
      ::InsertMenu( hMenu, 
         indexMenu++, 
         MF_STRING | MF_BYPOSITION, 
         idCmdFirst + IDM_CMD_PRINT, 
         s);
      s.LoadString(IDS_FINDTEXT);
      ::InsertMenu( hMenu, 
         indexMenu++, 
         MF_STRING | MF_BYPOSITION, 
         idCmdFirst + IDM_CMD_FINDTEXT, 
         s);
      return MAKE_HRESULT(SEVERITY_SUCCESS, 0, USHORT(IDM_CMD_LAST));
   }
   return MAKE_HRESULT(SEVERITY_SUCCESS, 0, USHORT(0));
}

STDMETHODIMP CHtmlBar::InvokeCommand(LPCMINVOKECOMMANDINFO lpcmi)
{
   ATLASSERT(lpcmi);
   switch( LOWORD(lpcmi->lpVerb) ) {
   case IDM_CMD_COPY:
      m_ctlText.Copy();
      break;
   case IDM_CMD_FINDTEXT:
      FindText();
      break;
   case IDM_CMD_PRINT:
      PrintText();
      break;
   default:
      return E_INVALIDARG;
   }
   return NOERROR;
}

STDMETHODIMP CHtmlBar::GetCommandString(UINT idCommand,
                                        UINT uFlags,
                                        LPUINT lpReserved,
                                        LPSTR lpszName,
                                        UINT uMaxNameLen)

{
   USES_CONVERSION;
   HRESULT hr = E_INVALIDARG;
   CString s;
   UINT Id;
   switch(uFlags) {
   case GCS_HELPTEXT:
      switch(idCommand) {
      case IDM_CMD_COPY:
         Id = IDS_TIP_COPY;
         break;
      case IDM_CMD_PRINT:
         Id = IDS_TIP_PRINT;
         break;
      case IDM_CMD_FINDTEXT:
         Id = IDS_TIP_FINDTEXT;
         break;
      default:
         return hr;
      }
      s.LoadString(Id);
      ::lstrcpy(lpszName, T2CA((LPCTSTR)s));
      hr = NOERROR;
      break;

   case GCS_VERB:
      switch(idCommand) {
      case IDM_CMD_COPY:
         Id = IDS_CMD_COPY;
         break;
      case IDM_CMD_PRINT:
         Id = IDS_CMD_PRINT;
         break;
      case IDM_CMD_FINDTEXT:
         Id = IDS_CMD_FINDTEXT;
         break;
      default:
         return hr;
      }
      s.LoadString(Id);
      ::lstrcpy(lpszName, T2CA((LPCTSTR)s));
      hr = NOERROR;
      break;

   case GCS_VALIDATE:
      hr = NOERROR;
      break;
   }
   return hr;
}


///////////////////////////////////////////////////////////////////////////
// Window creation

BOOL CHtmlBar::RegisterAndCreateWindow(void)
{
   // If the window doesn't exist yet, create it now.
   if( m_hWnd==NULL ) {
      // Can't create a child window without a parent.
      if( m_hwndParent==NULL ) return FALSE;

      // If the window class has not been registered, then do so.
      WNDCLASS wc;
      if( !::GetClassInfo(_Module.GetModuleInstance(), DB_CLASS_NAME, &wc) ) {
         ZeroMemory(&wc, sizeof(wc));
         wc.style          = CS_GLOBALCLASS;
         wc.lpfnWndProc    = (WNDPROC)WndProc;
         wc.cbClsExtra     = 0;
         wc.cbWndExtra     = 0;
         wc.hInstance      = _Module.GetModuleInstance();
         wc.hIcon          = NULL;
         wc.hCursor        = ::LoadCursor(NULL, IDC_ARROW);
         wc.hbrBackground  = (HBRUSH)COLOR_WINDOW;
         wc.lpszMenuName   = NULL;
         wc.lpszClassName  = DB_CLASS_NAME;
   
         if( !::RegisterClass(&wc) ) {
            // If RegisterClass fails, CreateWindow below will fail.
         }
      }

      // Create the window. The WndProc will set m_hWnd.
      CRect  rc;
      ::GetClientRect(m_hwndParent, &rc);
      HWND hWnd = ::CreateWindowEx( 0,
                      DB_CLASS_NAME,
                      NULL,
                      WS_CHILD | WS_CLIPSIBLINGS | WS_BORDER,
                      rc.left,
                      rc.top,
                      rc.Width(),
                      rc.Height(),
                      m_hwndParent,
                      NULL,
                      _Module.GetModuleInstance(),
                      (LPVOID)this);
      if( hWnd==NULL ) {
         ::MessageBeep((UINT)-1);
         return FALSE;
      }

      CRect rcDefault(0,0,10,10);
      m_ctlImages.Create(IDB_IMAGES, 16, 1, RGB(0,128,128));

      DWORD dwStyle;
      dwStyle = WS_CHILD | 
          WS_VISIBLE | 
          WS_BORDER |
          TVS_LINESATROOT |
          TVS_HASLINES |
          TVS_HASBUTTONS |
          TVS_SHOWSELALWAYS |
          TVS_DISABLEDRAGDROP;      
      m_ctlTree.Create(hWnd, &rcDefault, NULL, dwStyle, WS_EX_CLIENTEDGE);
      m_ctlTree.SetImageList(m_ctlImages, LVSIL_NORMAL);
      m_ctlTree.SetScrollTime(0L);

      dwStyle = WS_CHILD | 
         WS_VISIBLE | 
         WS_BORDER | 
         WS_HSCROLL |
         WS_VSCROLL |
         ES_MULTILINE |
         ES_NOHIDESEL |
         ES_READONLY;
      m_ctlText.Create(hWnd, &rcDefault, NULL, dwStyle, WS_EX_CLIENTEDGE);
      m_ctlText.SetFont((HFONT)::GetStockObject(DEFAULT_GUI_FONT));
      m_ctlText.SendMessage(EM_SETMARGINS, EC_LEFTMARGIN|EC_RIGHTMARGIN, MAKELONG(10, 10));
      m_ctlText.SendMessage(EM_AUTOURLDETECT, TRUE, 0L);
      m_ctlText.SetTargetDevice(NULL, 1);
      m_ctlText.SetEventMask(ENM_SELCHANGE | ENM_LINK);
      m_ctlText.LimitText(400000L);
      m_pRichEditOle = m_ctlText.GetOleInterface();
   }

   return (NULL != m_hWnd);
}


///////////////////////////////////////////////////////////////////////////
// Private method implementations

LRESULT CALLBACK CHtmlBar::WndProc(HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
   CHtmlBar  *pThis = (CHtmlBar*)::GetWindowLong(hWnd, GWL_USERDATA);
   switch( uMessage ) {
   case WM_NCCREATE:
      {
         LPCREATESTRUCT lpcs = (LPCREATESTRUCT)lParam;
         pThis = (CHtmlBar*)(lpcs->lpCreateParams);
         ::SetWindowLong(hWnd, GWL_USERDATA, (LONG)pThis);
         // Set the window handle
         pThis->m_hWnd = hWnd;
      }
      break;
   
   case WM_NOTIFY:
      return pThis->OnNotify(lParam);

   case WM_PAINT:
      return pThis->OnPaint();
   
   case WM_COMMAND:
      return pThis->OnCommand(wParam, lParam);
   
   case WM_SETFOCUS:
      return pThis->OnSetFocus();

   case WM_KILLFOCUS:
      return pThis->OnKillFocus();
   
   case WM_SIZE:
      return pThis->OnSize(lParam);

   default:
      if( uMessage==CFindReplaceDialog::GetFindReplaceMsg() ) {
         CFindReplaceDialog* pDlg = CFindReplaceDialog::GetNotifier(lParam);
         pThis->OnFindCmd(pDlg);
      }
      break;
   }
   return ::DefWindowProc(hWnd, uMessage, wParam, lParam);
}

LRESULT CHtmlBar::OnCommand(WPARAM wParam, LPARAM lParam)
{
   if( HIWORD(wParam)==EN_SETFOCUS ) OnEditGotFocus();
   if( HIWORD(wParam)==EN_KILLFOCUS ) OnEditLostFocus();
   return 0;
}

LRESULT CHtmlBar::OnNotify(LPARAM lParam)
{
   LPNMHDR pNmhdr = (LPNMHDR)lParam;
   ATLASSERT(pNmhdr);
   switch( pNmhdr->code ) {
   case TVN_SELCHANGED:
      {
         LPNM_TREEVIEW lpNMTreeView = (LPNM_TREEVIEW)lParam;
         TV_ITEM tvi = lpNMTreeView->itemNew;
         OnItemSelected(tvi);
      }
      break;
   case EN_SELCHANGE:
      {
         SELCHANGE *pSelChange = (SELCHANGE *)lParam;
         OnCursorMoved(pSelChange);
      }
      break;
   case EN_LINK: 
      {
         ENLINK *pLink = (ENLINK *)lParam;
         OnUrlClick(pLink);
      }
      break;
   }
   return 0;
}

LRESULT CHtmlBar::OnPaint(void)
{
   CPaintDC dc(m_hWnd);
   return 0;
}

LRESULT CHtmlBar::OnItemSelected(TV_ITEM &tvi)
{
   m_ctlText.SetWindowText(_T(""));
   m_bmp.Release();

   int idx = m_ctlTree.GetItemData(tvi.hItem);
   if( idx<=0 || idx>m_Entries.GetSize() ) return 0;
   CWaitCursor w;
   tEntry entry = m_Entries[idx-1];
   switch( entry.Type ) {
   case ITEM_ROOT:
      DisplayRoot(entry);
      break;
   case ITEM_DOCUMENT:
      DisplayDocument(entry);
      break;
   case ITEM_LINKS:
      DisplayLinks(entry);
      break;
   case ITEM_STYLESHEET:
      DisplayStyleSheet(entry);
      break;
   case ITEM_IMAGE:
      DisplayImage(entry);
      break;
   case ITEM_APPLET:
      DisplayApplet(entry);
      break;
   case ITEM_SCRIPT:
      DisplayScript(entry);
      break;
   }
   return 0;
}

LRESULT CHtmlBar::OnCursorMoved(SELCHANGE *pSelChange)
{
   CString s;
   int pos = pSelChange->chrg.cpMin;
   int line = m_ctlText.LineFromChar(pSelChange->chrg.cpMin);
   s.Format(IDS_STATUS_POS, pos+1, line+1);
   CComBSTR bstr(s);
   m_spWebBrowser2->put_StatusText(bstr);
   return 0;
}

LRESULT CHtmlBar::OnEditGotFocus()
{   
   SELCHANGE sel;
   m_ctlText.GetSel(sel.chrg);
   return OnCursorMoved(&sel);
}

LRESULT CHtmlBar::OnEditLostFocus()
{   
   if( m_spWebBrowser2==NULL ) return 0;
   CComBSTR bstr(L"");
   m_spWebBrowser2->put_StatusText(bstr);
   return 0;
}

LRESULT CHtmlBar::OnUrlClick(ENLINK *pLink)
{   
   if( m_spWebBrowser2==NULL ) return 0;
   if( pLink->msg!=WM_LBUTTONDOWN ) return 0;
   m_ctlText.SetSel(pLink->chrg);
   CComBSTR bstr;
   m_ctlText.GetSelTextBSTR(bstr.m_str); 
   CComVariant empty;
   m_spWebBrowser2->Navigate(bstr,&empty,&empty,&empty,&empty);
   return 0;
}

LRESULT CHtmlBar::OnSize(LPARAM lParam)
{
   int cx, cy;
   cx = LOWORD(lParam);
   cy = HIWORD(lParam);

   int XPAD = 4;
   int YPAD = 4;
   if( ::IsWindow(m_ctlTree) ) {
      //m_ctlTree.MoveWindow( XPAD, YPAD, cx-(XPAD*2), (cy/2)-(YPAD*2) );
      //m_ctlText.MoveWindow( XPAD, cy/2, cx-(XPAD*2), (cy/2)-(YPAD) );
      m_ctlTree.MoveWindow( XPAD, YPAD, (cx/3)-(XPAD*2), cy-(YPAD*2) );
      m_ctlText.MoveWindow( cx/3, YPAD, ((cx/3)*2)-(XPAD), cy-(YPAD*2) );
   }

   return 0;
}

void CHtmlBar::FocusChange(BOOL bFocus)
{
   m_bFocus = bFocus;
   // Inform the input object site that the focus has changed
   if( m_spSite!=NULL ) {
      m_spSite->OnFocusChangeIS((IDockingWindow*)this, bFocus);
   }
}

LRESULT CHtmlBar::OnSetFocus(void)
{
   FocusChange(TRUE);
   return 0;
}

LRESULT CHtmlBar::OnKillFocus(void)
{
   FocusChange(FALSE);
   return 0;
}


/////////////////////////////////////////////////////////////////////////////
// Browser connect

BOOL CHtmlBar::BrowserConnect(void)
{
   HRESULT hr;
   CComPtr<IConnectionPoint> spCP;

   // Receives the connection point for WebBrowser events
   hr = m_spCPC->FindConnectionPoint(
      DIID_DWebBrowserEvents2, 
      &spCP);
   if (FAILED(hr)) return FALSE;

   // Pass the event handlers to the container
   hr = spCP->Advise( 
      reinterpret_cast<IDispatch*>(this), 
      &m_dwCookie);
   return TRUE; 
}

BOOL CHtmlBar::BrowserDisconnect(void)
{
   HRESULT hr;
   CComPtr<IConnectionPoint> spCP;

   // Receives the connection point for WebBrowser events
   hr = m_spCPC->FindConnectionPoint(
      DIID_DWebBrowserEvents2, 
      &spCP);
   if (FAILED(hr)) return FALSE;

   // Stop getting event notifications
   hr = spCP->Unadvise(m_dwCookie);
   return TRUE;
}

HRESULT CHtmlBar::Invoke(DISPID dispidMember, 
                         REFIID riid, 
                         LCID lcid, 
                         WORD wFlags,
                         DISPPARAMS* pDispParams, 
                         VARIANT* pvarResult,
                         EXCEPINFO*  pExcepInfo,  
                         UINT* puArgErr)
// IDispatch callback that receives the event notifications
// from Internet Explorer.
{
   if( pDispParams==NULL ) return E_INVALIDARG;
   if( m_spWebBrowser2==NULL ) return E_FAIL;
 
   // Before Navigate
   switch( dispidMember ) {
   case DISPID_BEFORENAVIGATE:
      {
         ATLTRACE("IE: BeforeNavigate\n");
      }
      break;
   case DISPID_NAVIGATECOMPLETE:
      {
         // Before Navigate
         ATLTRACE("IE: NavigateComplete\n");
      }
      break;
   case DISPID_BEFORENAVIGATE2:
      {
         ATLTRACE("IE: BeforeNavigate2\n");
         // Clear tree
         if( m_ctlTree.GetCount()==1 ) break; // already in "wait" state
         RemoveTree();
         CString s;
         s.LoadString(IDS_LOADING);
         m_ctlTree.InsertItem(TVIF_TEXT|TVIF_STATE, s,0,0, TVIS_CUT,TVIS_CUT , 0, TVI_ROOT,NULL);
      }
      break;
   case DISPID_NAVIGATECOMPLETE2:
      {
         // Before Navigate
         ATLTRACE("IE: NavigateComplete2\n");
      }
      break;
   case DISPID_FRAMENAVIGATECOMPLETE:
      {
         // Frame navigation
         ATLTRACE("IE: FrameNavigateComplete\n");
      }
      break;
   case DISPID_DOWNLOADBEGIN:
      {
         // Download is about to begin
         ATLTRACE("IE: DownloadBegin\n");
      }
      break;
   case DISPID_DOWNLOADCOMPLETE:
      {
         // Download complete (to catch refresh and F5)
         ATLTRACE("IE: DownloadComplete\n");
         InitTree();
      }
      break;
   case DISPID_DOCUMENTCOMPLETE:
      {
         // Document complete
         ATLTRACE("IE: DocumentComplete\n");
         if( V_VT(&pDispParams->rgvarg[1])!=VT_DISPATCH ) break;
         // Need to check if its the top-level
         //CComPtr<IDispatch> spDisp( pDispParams->rgvarg[1].pdispVal );
         //if( m_spWebBrowser2.IsEqualObject(spDisp) ) {
         //   InitTree();
         //}
      }
      break;
   case DISPID_ONQUIT:
      {
         // Quit
         ATLTRACE("IE: OnQuit\n");
         RemoveTree();
      }
      break;
   }

   return S_OK;
}


/////////////////////////////////////////////////////////////////////////////
// Tree implementation

void CHtmlBar::InitTree()
{
   ATLTRACE("HtmlBar: InitTree\n");
   if( !::IsWindowVisible(m_hWnd) ) return;
   
   RemoveTree();

   CString s;
   DWORD dwStyle = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
   
   CComPtr<IDispatch> spDisp;
   HRESULT hr = m_spWebBrowser2->get_Document(&spDisp);
   if (FAILED(hr)) return;

   // Verify that what we get is a pointer to a IHTMLDocument2 
   // interface. To be sure, let's query for 
   // the IHTMLDocument2 interface (through smart pointers)
   CComQIPtr<IHTMLDocument2> spHTML ( spDisp );
   // The previous step is sufficient to keep Explorer aside
   // without an explicit check against the loader module
   if (spHTML!=NULL) {
      // Extract the elements of the document
      m_ctlTree.SetRedraw(FALSE);

      tEntry root; root.Type = ITEM_ROOT; root.spUnk = m_spWebBrowser2;
      m_Entries.Add(root);
      HTREEITEM hRoot;
      s.LoadString(IDS_TREEROOT);
      hRoot = m_ctlTree.InsertItem(dwStyle, s, 3,3, 0,0,m_Entries.GetSize(), TVI_ROOT, NULL);

#define ADDNODE(item, type, ptr, str, img1, img2) \
   tEntry item; item.Type = type; item.spUnk = ptr; \
   m_Entries.Add(item); \
   HTREEITEM h##item; \
   s.LoadString(str); \
   h##item = m_ctlTree.InsertItem(dwStyle, s, img1,img2, 0,0,m_Entries.GetSize(), hRoot, NULL);

      ADDNODE(Documents, ITEM_DOCUMENTS, NULL, IDS_TREEDOCUMENTS, 1,0);
      ADDNODE(Scripts, ITEM_SCRIPTS, NULL, IDS_TREESCRIPTS, 1,0);
      ADDNODE(Images, ITEM_IMAGES, NULL, IDS_TREEIMAGES, 1,0);
      ADDNODE(Applets, ITEM_APPLETS, NULL, IDS_TREEAPPLETS, 1,0);
      ADDNODE(StyleSheets, ITEM_STYLESHEETS, NULL, IDS_TREESTYLESHEETS, 1,0);
      ADDNODE(Links, ITEM_LINKS, spHTML, IDS_TREELINKS, 1,0);
#undef ADDNODE

      BuildTree_Frames(hRoot, hDocuments, hImages, hScripts, hLinks, hApplets, hStyleSheets, spHTML);

      m_ctlTree.Expand(hRoot);
      m_ctlTree.Expand(hDocuments);
      m_ctlTree.SelectItem(hRoot);
      m_ctlTree.SetRedraw(TRUE);
      m_ctlTree.Invalidate();
   }

}

void CHtmlBar::RemoveTree()
{
   ATLTRACE("HtmlBar: RemoveTree\n");
   m_ctlText.SetWindowText(NULL);
   m_ctlTree.DeleteAllItems();
   m_Entries.RemoveAll();
}

void CHtmlBar::BuildTree_Frames(HTREEITEM hRoot, 
                                HTREEITEM hDocuments, 
                                HTREEITEM hImages, 
                                HTREEITEM hScripts, 
                                HTREEITEM hLinks, 
                                HTREEITEM hApplets, 
                                HTREEITEM hStyleSheets, 
                                IHTMLDocument2 *pDoc)
// A helper method that builds the "Document" tree by recursively
// parsing the HTML hierachy (frames and documents)
{
   CComQIPtr<IHTMLDocument2> spHTML( pDoc );
   if( spHTML==NULL ) return;
   HRESULT hr;

   // Add known elements from this document to the other tree
   // branches (such as images, scripts etc etc).
   BuildTree_Images(hImages, spHTML);
   BuildTree_Links(hLinks, spHTML);
   BuildTree_Applets(hApplets, spHTML);
   BuildTree_Scripts(hScripts, spHTML);
   BuildTree_StyleSheets(hStyleSheets, spHTML);

   CComBSTR bstrTitle;
   hr = spHTML->get_title(&bstrTitle);
   if( FAILED(hr) ) return;
   if( bstrTitle.Length()==0 ) {
      bstrTitle.Empty();
      spHTML->get_URL(&bstrTitle);
   }

   tEntry document; document.Type = ITEM_DOCUMENT; document.spUnk = spHTML;
   m_Entries.Add(document);
   CString s = bstrTitle;
   DWORD dwStyle = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
   hDocuments = m_ctlTree.InsertItem(dwStyle, s, 2,2, 0,0,m_Entries.GetSize(), hDocuments, NULL);
   
   // Loop through all the frames in this document.
   // Recursively add new pages to this one...
   CComPtr<IHTMLFramesCollection2> spFrames;
   hr = spHTML->get_frames(&spFrames);
   if( FAILED(hr) ) return;
   long cnt;
   spFrames->get_length(&cnt);
   for( int i=0; i<cnt; i++ ) {
      CComVariant v(i);
      CComVariant res;
      spFrames->item(&v, &res);
      if( res.vt!=VT_DISPATCH ) continue;
      CComPtr<IDispatch> spDisp( res.pdispVal );
      CComQIPtr<IHTMLWindow2> spWindow( spDisp );
      if( spWindow!=NULL ) {
         CComQIPtr<IHTMLDocument2> spDoc;
         spWindow->get_document(&spDoc);
         if( spDoc!=NULL ) {
            BuildTree_Frames(hRoot, hDocuments, hImages, hScripts, hLinks, hApplets, hStyleSheets, spDoc);
         }
      }
   }

}

void CHtmlBar::BuildTree_Links(HTREEITEM hParent, IHTMLDocument2 *pDoc)
{
   // Dummy. we parse everything on display
}

void CHtmlBar::BuildTree_Scripts(HTREEITEM hParent, IHTMLDocument2 *pDoc)
{
   CComQIPtr<IHTMLDocument2> spHTML( pDoc );
   if( spHTML==NULL ) return;
   HRESULT hr;
   CComPtr<IHTMLElementCollection> spScripts;
   hr = spHTML->get_scripts(&spScripts);
   if( FAILED(hr) ) return;
   long cnt;
   spScripts->get_length(&cnt);
   CComVariant empty;
   for( int i=0; i<cnt; i++ ) {
      CComPtr<IDispatch> spDisp;
      spScripts->item(CComVariant(i), empty, &spDisp);
      CComQIPtr<IHTMLScriptElement> spEl( spDisp );
      if( spEl!=NULL ) {
         CComBSTR bstr;
         spEl->get_src(&bstr);
         // We display only imported script files!
         if( bstr.Length()>0 ) {
            CString s = bstr;
            tEntry script; script.Type = ITEM_SCRIPT; script.spUnk = spEl;
            m_Entries.Add(script);
            DWORD dwStyle = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
            m_ctlTree.InsertItem(dwStyle, s, 8,8, 0,0,m_Entries.GetSize(), hParent, NULL);
         }
      }
   }
}

void CHtmlBar::BuildTree_StyleSheets(HTREEITEM hParent, IHTMLDocument2 *pDoc)
{
   CComQIPtr<IHTMLDocument2> spHTML( pDoc );
   if( spHTML==NULL ) return;
   HRESULT hr;
   CComPtr<IHTMLStyleSheetsCollection> spSheets;
   hr = spHTML->get_styleSheets(&spSheets);
   if( FAILED(hr) ) return;
   long cnt;
   spSheets->get_length(&cnt);
   CComVariant empty;
   for( int i=0; i<cnt; i++ ) {
      CComPtr<IDispatch> spDisp;
      CComVariant idx(i);
      CComVariant res;
      spSheets->item(&idx, &res);
      if( res.vt!=VT_DISPATCH ) continue;
      CComQIPtr<IHTMLStyleSheet> spEl( res.pdispVal );
      if( spEl!=NULL ) {
         // The style sheet needs to be a downloadable (imported) entry
         // before we will show it.
         // (inline style sheets are displayed in the document as HTML)
         CComBSTR bstr;
         spEl->get_href(&bstr);
         if( bstr.Length()>0 ) {
            tEntry sheet; sheet.Type = ITEM_STYLESHEET; sheet.spUnk = spEl;
            m_Entries.Add(sheet);
            CString s = bstr;
            DWORD dwStyle = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
            m_ctlTree.InsertItem(dwStyle, s, 7,7, 0,0,m_Entries.GetSize(), hParent, NULL);
         }
      }
   }
}

void CHtmlBar::BuildTree_Images(HTREEITEM hParent, IHTMLDocument2 *pDoc)
{
   CComQIPtr<IHTMLDocument2> spHTML( pDoc );
   if( spHTML==NULL ) return;
   HRESULT hr;
   CComPtr<IHTMLElementCollection> spImages;
   hr = spHTML->get_images(&spImages);
   if( FAILED(hr) ) return;
   long cnt;
   spImages->get_length(&cnt);
   CComVariant empty;
   for( int i=0; i<cnt; i++ ) {
      CComPtr<IDispatch> spDisp;
      spImages->item(CComVariant(i), empty, &spDisp);
      CComQIPtr<IHTMLImgElement> spEl( spDisp );
      if( spEl!=NULL ) {
         CComBSTR bstr;
         spEl->get_src(&bstr);
         CString s = bstr;
         tEntry image; image.Type = ITEM_IMAGE; image.spUnk = spEl;
         m_Entries.Add(image);
         DWORD dwStyle = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
         m_ctlTree.InsertItem(dwStyle, s, 4,4, 0,0,m_Entries.GetSize(), hParent, NULL);
      }
   }
}

void CHtmlBar::BuildTree_Applets(HTREEITEM hParent, IHTMLDocument2 *pDoc)
{
   CComQIPtr<IHTMLDocument2> spHTML( pDoc );
   if( spHTML==NULL ) return;
   HRESULT hr;
   CComPtr<IHTMLElementCollection> spImages;
   hr = spHTML->get_applets(&spImages);
   if( FAILED(hr) ) return;
   long cnt;
   spImages->get_length(&cnt);
   CComVariant empty;
   for( int i=0; i<cnt; i++ ) {
      CComPtr<IDispatch> spDisp;
      spImages->item(CComVariant(i), empty, &spDisp);
      CComQIPtr<IHTMLElement> spEl( spDisp );
      if( spEl!=NULL ) {
         CComBSTR bstr;
         spEl->get_title(&bstr);
         if( bstr.Length()==0 ) {
            bstr.Empty();
            spEl->get_id(&bstr);
         }
         if( bstr.Length()==0 ) {
            bstr.Empty();
            bstr.LoadString(IDS_UNNAMED);
         }
         CString s = bstr;
         tEntry applet; applet.Type = ITEM_APPLET; applet.spUnk = spEl;
         m_Entries.Add(applet);
         DWORD dwStyle = TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
         m_ctlTree.InsertItem(dwStyle, s, 6,6, 0,0,m_Entries.GetSize(), hParent, NULL);
      }
   }
}


/////////////////////////////////////////////////////////////////////////////
// Display functions

#define TEXT2RTF(s) CPlainTextFormat::Format(s)
#define BSTR2RTF(s) CPlainTextFormat::Format(OLE2CT(s))

void CHtmlBar::DisplayRoot(tEntry &entry)
{
   USES_CONVERSION;
   CComQIPtr<IWebBrowser2> spWeb( entry.spUnk );
   if( spWeb==NULL ) return;

   CComBSTR bstrURL;
   CComBSTR bstrTitle;
   CComBSTR bstrType;
   spWeb->get_LocationName(&bstrTitle);
   spWeb->get_LocationURL(&bstrURL);
   spWeb->get_Type(&bstrType);

   CString txt;
   CString s;
   s.LoadString(IDS_RTF_SIMPLE_START);
   txt += s;
   s.Format(IDS_ROOTTEXT, BSTR2RTF(bstrTitle), BSTR2RTF(bstrURL), BSTR2RTF(bstrType));
   txt += s;
   s.LoadString(IDS_RTF_SIMPLE_END);
   txt += s;
   SetText(txt);

   CComBSTR bstr(L"HTML Source by Bjarke Viksoe");
   m_spWebBrowser2->put_StatusText(bstr);
}

void CHtmlBar::ParseLinks(IUnknown *pDoc, CString &s)
{
   USES_CONVERSION;
   ATLASSERT(pDoc);
   CComQIPtr<IHTMLDocument2> spHTML( pDoc );
   if( spHTML==NULL ) return;
   HRESULT hr;

   // Parse the HTML hierachy (frames and documents) for
   // all pages. Recursively going through all links.
   CComPtr<IHTMLFramesCollection2> spFrames;
   hr = spHTML->get_frames(&spFrames);
   if( FAILED(hr) ) return;
   long cnt;
   spFrames->get_length(&cnt);
   for( int i=0; i<cnt; i++ ) {
      CComVariant v(i);
      CComVariant res;
      spFrames->item(&v, &res);
      if( res.vt!=VT_DISPATCH ) continue;
      CComPtr<IDispatch> spDisp( res.pdispVal );
      CComQIPtr<IHTMLWindow2> spWindow( spDisp );
      if( spWindow!=NULL ) {
         CComQIPtr<IHTMLDocument2> spDoc;
         spWindow->get_document(&spDoc);
         if( spDoc!=NULL ) {
            ParseLinks(spDoc, s);
         }
      }
   }

   // Add a title for each html page
   CComBSTR bstrTitle;
   hr = spHTML->get_title(&bstrTitle);
   if( FAILED(hr) ) return;
   if( bstrTitle.Length()==0 ) {
      bstrTitle.Empty();
      spHTML->get_URL(&bstrTitle);
   }
   CString txt;
   txt.Format(IDS_LINKCAPTION, BSTR2RTF(bstrTitle));
   s += txt;

   CComPtr<IHTMLElementCollection> spLinks;
   hr = spHTML->get_links(&spLinks);
   if( FAILED(hr) ) return;
   // Add the links
   spLinks->get_length(&cnt);
   CComVariant empty;
   for( i=0; i<cnt; i++ ) {
      CComPtr<IDispatch> spDisp;
      spLinks->item(CComVariant(i), empty, &spDisp);
      CComQIPtr<IHTMLAnchorElement> spEl( spDisp );
      if( spEl!=NULL ) {
         CComBSTR bstr;
         spEl->get_href(&bstr);
         CString txt;
         txt.Format(IDS_LINKTEXT, BSTR2RTF(bstr));
         s += txt;
      }
   }
}

void CHtmlBar::DisplayLinks(tEntry &entry)
{
   CString txt;
   CString s;
   s.LoadString(IDS_RTF_SIMPLE_START);
   txt += s;
   s.Empty();
   ParseLinks(entry.spUnk, s);
   txt += s;
   s.LoadString(IDS_RTF_SIMPLE_END);
   txt += s;
   SetText(txt);
}

void CHtmlBar::DisplayDocument(tEntry &entry)
{
   CComQIPtr<IHTMLDocument2> spHTML( entry.spUnk );
   if( spHTML==NULL ) return;
   
   CComBSTR bstr;
   HRESULT hr;

   // Find the HTML element and extract outerHTML attribute value
   CComPtr<IHTMLElementCollection> spColl;
   hr = spHTML->get_all(&spColl);
   if( FAILED(hr) ) return;
   CComPtr<IDispatch> spDisp;
   hr = spColl->tags(CComVariant(L"HTML"), &spDisp);
   if( FAILED(hr) ) return;
   if( spDisp==NULL ) return;
   CComQIPtr<IHTMLElementCollection> spHTMLs( spDisp );
   spDisp.Release();
   if( spHTMLs==NULL ) return;
   long cnt;
   spHTMLs->get_length(&cnt);
   if( cnt<=0 ) return;
   CComVariant empty;
   hr = spHTMLs->item(CComVariant(0L), empty, &spDisp);
   if( FAILED(hr) ) return;
   CComQIPtr<IHTMLElement> spEl( spDisp );
   spDisp.Release();
   if( spEl==NULL ) return;

   hr = spEl->get_outerHTML(&bstr);
   if( FAILED(hr) ) return;

   CString txt;
   CString s;
   s.LoadString(IDS_RTF_HTML_START);
   txt += s;
   s = bstr;
   txt += CHtmlFormat::Format(s);
   s.LoadString(IDS_RTF_HTML_END);
   txt += s;
   SetText(txt);
}

BOOL CHtmlBar::ParseScript(IDebugApplicationNode *pNode, CString &sScriptName, CString &sContents)
{
   if( pNode==NULL ) return FALSE;
   CComPtr<IDebugApplicationNode> spNode( pNode );
   HRESULT hr;

   CComQIPtr<IDebugDocumentInfo> spDocInfo( spNode );
   if( spDocInfo!=NULL ) {
      CComBSTR bstrName;
      hr = spDocInfo->GetName(DOCUMENTNAMETYPE_URL, &bstrName);
      if( SUCCEEDED(hr) ) {
         CString sName(bstrName);
         sName.MakeLower();
         sName.Replace(_T('\\'), _T('/'));
         if( sName.Find(sScriptName)>=0 ) {
            // Match!!!
            CComQIPtr<IDebugDocumentProvider> spDocProvider( spNode );
            if( spDocProvider!=NULL ) {
               CComPtr<IDebugDocument> spDoc;
               hr = spDocProvider->GetDocument(&spDoc);
               if( SUCCEEDED(hr) ) {
                  CComQIPtr<IDebugDocumentText> spDocText(spDoc);
                  if( spDocText!=NULL ) {
                     ULONG lLines, lChars;
                     hr = spDocText->GetSize(&lLines, &lChars);
                     if( FAILED(hr) ) return FALSE;
                     lChars += 10;
                     WCHAR *pStr = new WCHAR[lChars+1];
                     if( pStr==NULL ) return FALSE;
                     SOURCE_TEXT_ATTR *pAttribs = new SOURCE_TEXT_ATTR[lChars+1];
                     if( pAttribs==NULL ) return FALSE;
                     ULONG dwRetrieved = 0;
                     hr = spDocText->GetText(0, pStr, pAttribs, &dwRetrieved, lChars);
                     if( SUCCEEDED(hr) ) {
                        *(pStr + dwRetrieved) = L'\0';
                        sContents = pStr;
                     }
                     delete [] pStr;
                     delete [] pAttribs;
                     return TRUE;
                  }
               }
            }
         }
      }
   }

   // Parse all sub nodes
   CComPtr<IEnumDebugApplicationNodes> spNodes;
   hr = spNode->EnumChildren(&spNodes);
   if( SUCCEEDED(hr) ) {
      CComPtr<IDebugApplicationNode> spChild;
      DWORD dwFetched;
      spNodes->Next(1, &spChild, &dwFetched);
      while( dwFetched==1 ) {
         BOOL res = ParseScript(spChild, sScriptName, sContents);
         if( res ) return TRUE;
         spChild.Release();
         spNodes->Next(1, &spChild, &dwFetched);
      }
   }
   return FALSE;
}

BOOL CHtmlBar::ParseCachedScript(CString& sScriptName, CString& sContents)
{
   // Load script from cache internet area
   DWORD dwEntrySize = 0;
   ::RetrieveUrlCacheEntryStream(sScriptName, NULL, &dwEntrySize, FALSE, 0);
   LPINTERNET_CACHE_ENTRY_INFO lpCacheEntry;
   HANDLE hStream;
   lpCacheEntry = LPINTERNET_CACHE_ENTRY_INFO(new char[dwEntrySize]);
   if( !(hStream = ::RetrieveUrlCacheEntryStream(sScriptName, lpCacheEntry, &dwEntrySize, TRUE, 0)) ) {
      return FALSE;
   }
   DWORD dwStreamSize = lpCacheEntry->dwSizeLow;
   LPBYTE lpszOut = new BYTE[dwStreamSize + 1];
   if( !::ReadUrlCacheEntryStream(hStream,0, (LPVOID) lpszOut, &dwStreamSize, 0) ) {
      delete [] lpszOut;
      delete [] lpCacheEntry;
      ::UnlockUrlCacheEntryStream(hStream,0);
      return FALSE;
   }

   lpszOut[ dwStreamSize ] = '\0';
   sContents = (LPCTSTR) lpszOut;

   delete [] lpszOut;
   delete [] lpCacheEntry;
   ::UnlockUrlCacheEntryStream(hStream,0);
  
   return TRUE;
}

void CHtmlBar::DisplayScript(tEntry &entry)
{
   CComQIPtr<IHTMLScriptElement> spScript( entry.spUnk );
   if( spScript==NULL ) return;
   HRESULT hr;

   // Get the script url.
   // NOTE: We use this to locate the script in the "Running Documents"
   //       list of the Active Script Debugger.
   //       I have found no better way to do this!
   CComBSTR bstr;
   hr = spScript->get_src(&bstr);
   if( FAILED(hr) ) return;
   CString sTitle(bstr);
   sTitle.MakeLower();
   sTitle.Replace(_T('\\'), _T('/'));

   // Determine script type (needed for proper syntax coloring)
   CComBSTR bstrType;
   hr = spScript->get_type(&bstrType);
   if( FAILED(hr) ) return;
   if( bstrType.Length()==0 ) {
      bstrType.Empty();
      CComQIPtr<IHTMLElement> spEl( entry.spUnk );
      if( spEl==NULL ) return;
      hr = spEl->get_language(&bstrType);
      if( FAILED(hr) ) return;
   }
   CString sType(bstrType);
   // Often the 'type' attribute is empty. We need
   // to make up a type...
   if( sType.GetLength()==0 ) {
      // First try with the file extension
      CString sFile( sTitle );
      sFile.MakeLower();
      if( sFile.Find(_T(".js"))>0 ) sType=_T("JScript");
      if( sFile.Find(_T(".vb"))>0 ) sType=_T("VBScript");
      if( sFile.Find(_T(".vbs"))>0 ) sType=_T("VBScript");
   }
   if( sType.GetLength()==0 ) sType=_T("JavaScript"); // default is JavaScript for all browsers
   sType.MakeLower();

   CString sContents;

   // To locate the script code, I didn't find any other
   // way than to query the Active Script Manager.
   // This seems to be installed with a "Script Debugger", so on many
   // machines this will fail!
   CComPtr<IMachineDebugManager> spManager;
   CLSID clsid;
   ::CLSIDFromString(L"{0C0A3666-30C9-11D0-8F20-00805F2CD064}", &clsid); // CLSID_MachineDebugManager
   HRESULT HrManager = spManager.CoCreateInstance(clsid);

   if( SUCCEEDED(HrManager) && sContents.GetLength()==0 ) {
      // Scan all "Running Documents"...
      CComPtr<IEnumRemoteDebugApplications> spApps;
      hr = spManager->EnumApplications(&spApps);
      if( FAILED(hr) ) return;
      if( spApps==NULL ) return;
      CComPtr<IRemoteDebugApplication> spApp;
      DWORD dwFetched = 0;
      spApps->Next(1, &spApp, &dwFetched);
      while( dwFetched==1 ) {
         CComPtr<IDebugApplicationNode> spRoot;
         hr = spApp->GetRootNode(&spRoot);
         if( SUCCEEDED(hr) && spRoot!=NULL ) {
            if( ParseScript(spRoot, sTitle, sContents)==TRUE ) break;
         }
         spApp.Release();
         spApps->Next(1, &spApp, &dwFetched);
      }
   }
   // Actually it is possible to locate the scripting code, if it
   // is in the internal IE cache. So look for it there as well...
   if( sContents.GetLength()==0 ) {
      ParseCachedScript(sTitle, sContents);
   }
   if( FAILED(HrManager) && sContents.GetLength()==0 ) {
      sContents.LoadString(IDS_ERR_SCRIPTNOTAVAIL);
   }
   // Someone was kind enough to deliver some scripting text.
   // Show it!!!
   if( sContents.GetLength()>0 ) {
      CString txt;
      CString s;
      s.LoadString(IDS_RTF_SCRIPT_START);
      txt += s;
      if( sType.Find(_T("ecma"))>=0 
          || sType.Find(_T("javascript"))>=0 
          || sType.Find(_T("jscript"))>=0 ) {
         txt += CJScriptFormat::Format(sContents);
      }
      else if( sType.Find(_T("vbscript"))>=0 ) {
         txt += CVBScriptFormat::Format(sContents);
      }
      else {
         txt += CScriptFormat::Format(sContents);
      }
      s.LoadString(IDS_RTF_SCRIPT_END);
      txt += s;
      SetText(txt);
   }
   else {
      // Script not found
      CString txt;
      CString s;
      s.LoadString(IDS_RTF_SIMPLE_START);
      txt += s;
      s.LoadString(IDS_ERR_SCRIPTLOAD);
      txt += s;
      s.LoadString(IDS_RTF_SIMPLE_END);
      txt += s;
      SetText(txt);
   }
}

void CHtmlBar::DisplayImage(tEntry &entry)
{
   USES_CONVERSION;
   CComQIPtr<IHTMLImgElement> spEl( entry.spUnk );
   if( spEl==NULL ) return;

   CComBSTR bstrTitle;
   CComBSTR bstrURL;
   CComBSTR bstrAlt;
   long lWidth;
   long lHeight;
   CComBSTR bstrType;
   CComBSTR bstrFileSize;
   spEl->get_name(&bstrTitle);
   if( bstrTitle.Length()==0 ) {
      bstrTitle.Empty();
      bstrTitle.LoadString(IDS_UNNAMED);
   }
   spEl->get_src(&bstrURL);
   spEl->get_alt(&bstrAlt);
   spEl->get_width(&lWidth);
   spEl->get_height(&lHeight);
   spEl->get_mimeType(&bstrType);
   spEl->get_fileSize(&bstrFileSize);

   CString txt;
   CString s;
   s.LoadString(IDS_RTF_SIMPLE_START);
   txt += s;
   s.Format(IDS_IMAGETEXT, BSTR2RTF(bstrTitle), BSTR2RTF(bstrURL), BSTR2RTF(bstrAlt), lWidth, lHeight, BSTR2RTF(bstrType), BSTR2RTF(bstrFileSize));
   txt += s;
   s.LoadString(IDS_RTF_SIMPLE_END);
   txt += s;
   SetText(txt);

   // Download and display image   
   if( bstrURL != NULL && m_pRichEditOle != NULL ) {
      USES_CONVERSION;
      IPicture* pPicture = NULL;
      if( SUCCEEDED( ::OleLoadPicturePath(bstrURL, NULL, 0, RGB(255,255,255), IID_IPicture, (LPVOID*) &pPicture) ) ) {
         OLE_HANDLE hPic = NULL;
         pPicture->get_Handle(&hPic);
         HBITMAP hBitmap = (HBITMAP) ::CopyImage((HANDLE) hPic, IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
         if( m_bmp.CreateFromBitmap(m_hWnd, m_pRichEditOle, hBitmap) ) {
            m_bmp.Insert(m_ctlText, m_pRichEditOle);
         }
         pPicture->Release();
      }
   }

   m_ctlText.SetSel(0,0);
}

void CHtmlBar::DisplayApplet(tEntry &entry)
{
   USES_CONVERSION;
   CComQIPtr<IHTMLElement> spEl( entry.spUnk );
   if( spEl==NULL ) return;

   CComBSTR bstrTitle;
   CComBSTR bstrId;
   CComBSTR bstrAlt;
   CComBSTR bstrCode;
   CComBSTR bstrCodeBase;

   spEl->get_id(&bstrId);
   if( bstrId.Length()==0 ) {
      bstrId.Empty();
      bstrId.LoadString(IDS_UNNAMED);
   }
   spEl->get_title(&bstrTitle);
   spEl->get_className(&bstrAlt);
   // Get code/class/classid from attribute list
   if( bstrCode.Length()==0 ) {
      bstrCode.Empty();
      CComVariant vCode;
      spEl->getAttribute(CComBSTR(L"CODE"), 0, &vCode);
      if( vCode.vt==VT_BSTR ) bstrCode = V_BSTR(&vCode);
   }
   if( bstrCode.Length()==0 ) {
      bstrCode.Empty();
      CComVariant vCode;
      spEl->getAttribute(CComBSTR(L"CLSID"), 0, &vCode);
      if( vCode.vt==VT_BSTR ) bstrCode = V_BSTR(&vCode);
   }
   if( bstrCode.Length()==0 ) {
      bstrCode.Empty();
      CComVariant vCode;
      spEl->getAttribute(CComBSTR(L"CLASSID"), 0, &vCode);
      if( vCode.vt==VT_BSTR ) bstrCode = V_BSTR(&vCode);
   }
   CComVariant vCodeBase;
   spEl->getAttribute(CComBSTR(L"CODEBASE"), 0, &vCodeBase);
   if( vCodeBase.vt==VT_BSTR ) bstrCodeBase = V_BSTR(&vCodeBase);

   CString txt;
   CString s;
   s.LoadString(IDS_RTF_SIMPLE_START);
   txt += s;
   s.Format(IDS_APPLETTEXT, 
      BSTR2RTF(bstrId), 
      BSTR2RTF(bstrTitle), 
      BSTR2RTF(bstrAlt), 
      BSTR2RTF(bstrCode), 
      BSTR2RTF(bstrCodeBase));
   txt += s;
   s.LoadString(IDS_RTF_SIMPLE_END);
   txt += s;
   SetText(txt);
}

void CHtmlBar::DisplayStyleSheet(tEntry &entry)
{
   CComQIPtr<IHTMLStyleSheet> spEl( entry.spUnk );
   if( spEl==NULL ) return;

   CComBSTR bstr;
   HRESULT hr;
   hr = spEl->get_cssText(&bstr);
   if( FAILED(hr) ) return;

   CString txt;
   CString s;
   s.LoadString(IDS_RTF_HTML_START);
   txt += s;
   s = bstr;
   txt += CCssFormat::Format(s);
   s.LoadString(IDS_RTF_HTML_END);
   txt += s;
   SetText(txt);
}

void CHtmlBar::FindText()
{
   BSTR bstr = NULL;
   m_ctlText.GetSelTextBSTR(bstr);
   CString sFindText = bstr;
   ::SysFreeString(bstr);
   
   if(m_pFindDlg != NULL) {
      ATLASSERT(::IsWindow(m_pFindDlg->m_hWnd));
      m_pFindDlg->SetFocus();
      return;
   }
   m_pFindDlg = new CFindReplaceDialog;
   if(m_pFindDlg == NULL) {
      _ASSERTE(FALSE);
      ::MessageBeep((UINT)-1);
      return;
   }
   DWORD dwFlags = FR_HIDEUPDOWN;
   if(!m_pFindDlg->Create(TRUE, sFindText, NULL, dwFlags, m_hWnd)) {
      delete m_pFindDlg;
      m_pFindDlg = NULL;
      ::MessageBeep((UINT)-1);
      return;
   }
   m_pFindDlg->ShowWindow(SW_NORMAL);
}

void CHtmlBar::OnFindCmd(CFindReplaceDialog* pDlg)
{
   if( pDlg==NULL ) return;
   ATLASSERT(pDlg==m_pFindDlg);
   
   // If the dialog is closing...
   if( pDlg->IsTerminating() ) {
      ATLASSERT(m_pFindDlg);
      m_pFindDlg = NULL; // NULL it!
      return;
   }
   if( !pDlg->FindNext() ) return;
	
	// Look for the next instance...
   TCHAR szFind[128];
   ::lstrcpyn(szFind, pDlg->m_fr.lpstrFindWhat, sizeof(szFind)/sizeof(TCHAR));
	BOOL bMatchCase = pDlg->MatchCase();
	BOOL bWholeWord = pDlg->MatchWholeWord();
	DWORD dwFlags = FR_DOWN;
	if( bMatchCase ) dwFlags |= FR_MATCHCASE;
	if( bWholeWord ) dwFlags |= FR_WHOLEWORD;

	CHARRANGE chrg;
	m_ctlText.GetSel(chrg);
	FINDTEXTEX ft;
	ft.chrg.cpMin = chrg.cpMax;
	ft.chrg.cpMax = -1;
	ft.lpstrText = szFind;
   if( m_ctlText.FindText(dwFlags, ft) == -1 ) {
      ::MessageBeep((UINT)-1);
      return;
   }
	m_ctlText.SetSel(ft.chrgText);
}

void CHtmlBar::PrintText()
{
   ATLASSERT(::IsWindow(m_ctlText.m_hWnd));
   
   HWND hWndRichEdit = m_ctlText.m_hWnd;

   // Initialize the PRINTDLG structure.
   PRINTDLG pd = { 0 };
   pd.lStructSize = sizeof(PRINTDLG);
   pd.hwndOwner = hWndRichEdit;
   pd.hDevMode = (HANDLE)NULL;
   pd.hDevNames = (HANDLE)NULL;
   pd.nFromPage = 0;
   pd.nToPage = 0;
   pd.nMinPage = 0;
   pd.nMaxPage = 0;
   pd.nCopies = 0;
   pd.hInstance = _Module.GetModuleInstance();
   pd.Flags = PD_RETURNDC | PD_NOPAGENUMS | PD_NOSELECTION | PD_PRINTSETUP;
   pd.lpfnSetupHook = (LPSETUPHOOKPROC)(FARPROC)NULL;
   pd.lpSetupTemplateName = (LPTSTR)NULL;
   pd.lpfnPrintHook = (LPPRINTHOOKPROC)(FARPROC)NULL;
   pd.lpPrintTemplateName = (LPTSTR)NULL;

   // Get the printer DC.
   if( ::PrintDlg(&pd)==TRUE ) {
      // Fill out the FORMATRANGE structure for the RTF output.
      FORMATRANGE fr = { 0 };
      fr.hdc = fr.hdcTarget = pd.hDC; // HDC
      fr.chrg.cpMin = 0; // print 
      fr.chrg.cpMax = -1; // entire contents
      fr.rcPage.left = fr.rcPage.top = 0;
      fr.rcPage.bottom = ::GetDeviceCaps(pd.hDC, PHYSICALHEIGHT);
      fr.rcPage.right = ::GetDeviceCaps(pd.hDC, PHYSICALWIDTH);

      fr.rcPage.right = MulDiv(fr.rcPage.right, 1440, ::GetDeviceCaps(pd.hDC, LOGPIXELSX));
      fr.rcPage.bottom = MulDiv(fr.rcPage.bottom, 1440, ::GetDeviceCaps(pd.hDC, LOGPIXELSY));

      RECT rcMargin = {1000,1000,1000,1000};
      fr.rc = fr.rcPage;
      // Convert from 1/1000" to twips
      fr.rc.left += MulDiv(rcMargin.left, 1440, 1000);
      fr.rc.right -= MulDiv(rcMargin.right, 1440, 1000);
      fr.rc.top += MulDiv(rcMargin.top, 1440, 1000);
      fr.rc.bottom -= MulDiv(rcMargin.bottom, 1440, 1000);

      // Fill out the DOCINFO structure.
      DOCINFO docInfo = { 0 };
      docInfo.cbSize = sizeof(DOCINFO);
      docInfo.lpszDocName = _T("(Html Source)");
      docInfo.lpszOutput = NULL;

      // Be sure that the printer DC is in text mode.
      ::SetMapMode (pd.hDC, MM_TEXT);

      ::StartDoc(pd.hDC, &docInfo);
      ::StartPage(pd.hDC);

      LONG lTextOut, lTextAmt;
      lTextOut = 0;
      lTextAmt = ::SendMessage(hWndRichEdit, WM_GETTEXTLENGTH, 0, 0);
      BOOL bSuccess = TRUE;

      while( bSuccess && (lTextOut < lTextAmt) ) {
         lTextOut = ::SendMessage(hWndRichEdit, EM_FORMATRANGE, FALSE, (LPARAM)&fr);
         bSuccess = ::SendMessage(hWndRichEdit, EM_DISPLAYBAND, 0, (LPARAM)&fr.rc);
         if( fr.chrg.cpMin==lTextOut ) bSuccess = FALSE; // HACK: Didn't get anywhere?
         if( lTextOut < lTextAmt ) {
            ::EndPage(pd.hDC);
            ::StartPage(pd.hDC); 
            fr.chrg.cpMin = lTextOut;
            fr.chrg.cpMax = -1;
         }
      }

      // Reset the formatting of the rich edit control.
      ::SendMessage(hWndRichEdit, EM_FORMATRANGE, FALSE, (LPARAM)NULL);

      // Finish the document.
      ::EndPage(pd.hDC);
      ::EndDoc(pd.hDC);

      // Delete the printer DC.
      ::DeleteDC(pd.hDC);
   }
}

DWORD CALLBACK CHtmlBar::StreamReadCallback(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG FAR *pcb)
{
   _ASSERTE(dwCookie==0);
   _ASSERTE(pcb!=NULL);

   long lNeeded;
#ifndef _UNICODE
   LPCSTR pStr = m_sStream;
   pStr += m_lStreamIndex;
   lNeeded = (LONG)strlen(pStr) < cb ? strlen(pStr) : cb;
   strncpy( (LPSTR)pbBuff, pStr, lNeeded );
#else
#error Unicode not implemented
#endif
   *pcb = lNeeded;
   m_lStreamIndex += lNeeded;
   return 0;
}

void CHtmlBar::SetText(LPCTSTR s)
{
   if( (LONG)_tcslen(s)>m_ctlText.GetLimitText() ) {
      ATLASSERT("text too large"==NULL);
      ::MessageBeep((UINT)-1);
      return;
   }
   // To set the RTF text we actually need to stream it into
   // the control. SetWindowText() only works properly on W2k.
   // NOTE: We pass static variables here to the callback
   //       just to make life easier.
   m_sStream = s;
   m_lStreamIndex = 0;
   EDITSTREAM es;
   es.dwCookie = 0L;
   es.dwError = 0;
   es.pfnCallback = StreamReadCallback;
   m_ctlText.StreamIn(SF_RTF, es);
}
