// HtmlExplorerBar.h : Declaration of the CHtmlBar

#ifndef __HTMLBAR_H_
#define __HTMLBAR_H_
//
// Written by Bjarke Viksoe (bjarke@viksoe.dk)
// Copyright (c) 2000.
//
// This code may be used in compiled form in any way you desire. This
// file may be redistributed unmodified by any means PROVIDING it is 
// not sold for profit without the authors written consent, and 
// providing that this notice and the authors name is included. 
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability if it causes any damage to you or your
// computer whatsoever. It's free, so don't hassle me about it.
//
// Beware of bugs.
//

#include "resource.h"       // main symbols
#include "stdafx.h"

#include "RtfBitmap.h"


/////////////////////////////////////////////////////////////////////////////
// Entry

typedef enum { 
   ITEM_UNKNOWN = 0,
   ITEM_ROOT,
   ITEM_DOCUMENTS,
   ITEM_LINKS,
   ITEM_APPLETS,
   ITEM_STYLESHEETS,
   ITEM_IMAGES,
   ITEM_SCRIPTS,
   ITEM_DOCUMENT,
   ITEM_LINK,
   ITEM_APPLET,
   ITEM_IMAGE,
   ITEM_STYLESHEET,
   ITEM_SCRIPT
} TYPE;

typedef struct tagEntry
{
   TYPE Type;
   CComPtr<IUnknown> spUnk;
} tEntry;


/////////////////////////////////////////////////////////////////////////////
// CHtmlBar

#define DB_CLASS_NAME (TEXT("HtmlDeskBandClass"))

#define MIN_SIZE_X   10
#define MIN_SIZE_Y   10

#define IDM_CMD_COPY 0
#define IDM_CMD_PRINT 1
#define IDM_CMD_FINDTEXT 2

#define IDM_CMD_LAST 3


class ATL_NO_VTABLE CHtmlBar : 
   public CComObjectRootEx<CComSingleThreadModel>,
   public CComCoClass<CHtmlBar, &CLSID_HtmlBar>,
   public IDeskBand,
   public IInputObject,
   public IObjectWithSite,
   public IContextMenu,
   public IPersistStream,
   public IDispatchImpl<IHtmlBar, &IID_IHtmlBar, &LIBID_HTMLBARLib>
{
public:
   HRESULT FinalConstruct();
   void FinalRelease();

DECLARE_REGISTRY_RESOURCEID(IDR_HTMLBAR)
DECLARE_NOT_AGGREGATABLE(CHtmlBar)
DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CHtmlBar)
   COM_INTERFACE_ENTRY(IHtmlBar)
   COM_INTERFACE_ENTRY(IDispatch)
   COM_INTERFACE_ENTRY(IDeskBand)
   COM_INTERFACE_ENTRY(IOleWindow)
   COM_INTERFACE_ENTRY(IDockingWindow)
   COM_INTERFACE_ENTRY(IInputObject)
   COM_INTERFACE_ENTRY(IObjectWithSite)
   COM_INTERFACE_ENTRY(IPersistStream)
   COM_INTERFACE_ENTRY(IPersist)
   COM_INTERFACE_ENTRY(IContextMenu)
END_COM_MAP()

BEGIN_CATEGORY_MAP(CHtmlBar)
   IMPLEMENTED_CATEGORY(CATID_CommBand)
END_CATEGORY_MAP()

public:
   // IOleWindow methods
   STDMETHOD(GetWindow)(HWND*);
   STDMETHOD(ContextSensitiveHelp)(BOOL);

   // IDispatch
   STDMETHOD(Invoke)(DISPID, REFIID, LCID, WORD, DISPPARAMS*, 
                     VARIANT*, EXCEPINFO*, UINT*);

   // IDockingWindow methods
   STDMETHOD(ShowDW)(BOOL fShow);
   STDMETHOD(CloseDW)(DWORD dwReserved);
   STDMETHOD(ResizeBorderDW)(LPCRECT prcBorder, IUnknown* punkToolbarSite, BOOL fReserved);

   // IDeskBand methods
   STDMETHOD(GetBandInfo)(DWORD, DWORD, DESKBANDINFO*);

   // IInputObject methods
   STDMETHOD(UIActivateIO)(BOOL, LPMSG);
   STDMETHOD(HasFocusIO)(void);
   STDMETHOD(TranslateAcceleratorIO)(LPMSG);

   // IObjectWithSite methods
   STDMETHOD(SetSite)(IUnknown*);
   STDMETHOD(GetSite)(REFIID, LPVOID*);

   // IPersistStream methods
   STDMETHOD(GetClassID)(LPCLSID);
   STDMETHOD(IsDirty)(void);
   STDMETHOD(Load)(LPSTREAM);
   STDMETHOD(Save)(LPSTREAM, BOOL);
   STDMETHOD(GetSizeMax)(ULARGE_INTEGER*);

   // IContextMenu methods
   STDMETHOD(QueryContextMenu)(HMENU, UINT, UINT, UINT, UINT);
   STDMETHOD(InvokeCommand)(LPCMINVOKECOMMANDINFO);
   STDMETHOD(GetCommandString)(UINT, UINT, LPUINT, LPSTR, UINT);

protected:
   static LRESULT CALLBACK WndProc(HWND hWnd, UINT uMessage, WPARAM wParam, LPARAM lParam);
   void FocusChange(BOOL);
   LRESULT OnKillFocus(void);
   LRESULT OnSetFocus(void);
   LRESULT OnPaint(void);
   LRESULT OnCommand(WPARAM wParam, LPARAM lParam);
   LRESULT OnNotify(LPARAM lParam);
   LRESULT OnItemSelected(TV_ITEM &tvi);
   LRESULT OnSize(LPARAM);
   LRESULT OnCursorMoved(SELCHANGE *pSelChange);
   LRESULT OnEditGotFocus();
   LRESULT OnEditLostFocus();
   LRESULT OnUrlClick(ENLINK *pLink);
   BOOL RegisterAndCreateWindow(void);
   //
   BOOL BrowserConnect(void);
   BOOL BrowserDisconnect(void);
   //
   void InitTree();
   void RemoveTree();
   void BuildTree_Frames(HTREEITEM hParent, HTREEITEM hDocuments, 
                         HTREEITEM hImages, HTREEITEM hScripts, 
                         HTREEITEM hLinks, HTREEITEM hApplets, 
                         HTREEITEM hStyleSheets, 
                         IHTMLDocument2 *pDoc);
   void BuildTree_Applets(HTREEITEM hParent, IHTMLDocument2 *pDoc);
   void BuildTree_Links(HTREEITEM hParent, IHTMLDocument2 *pDoc);
   void BuildTree_Images(HTREEITEM hParent, IHTMLDocument2 *pDoc);
   void BuildTree_Scripts(HTREEITEM hParent, IHTMLDocument2 *pDoc);
   void BuildTree_StyleSheets(HTREEITEM hParent, IHTMLDocument2 *pDoc);
   //
   void DisplayRoot(tEntry &entry);
   void DisplayDocument(tEntry &entry);
   void DisplayLinks(tEntry &entry);
   void DisplayStyleSheet(tEntry &entry);
   void DisplayScript(tEntry &entry);
   void DisplayImage(tEntry &entry);
   void DisplayApplet(tEntry &entry);
   void ParseLinks(IUnknown *pDoc, CString &s);
   BOOL ParseScript(IDebugApplicationNode *pNode, CString &sScriptName, CString &sContents);
   BOOL ParseCachedScript(CString& sScriptName, CString& sContents);
   void PrintText();
   void FindText();
   void OnFindCmd(CFindReplaceDialog* pDlg);
   void SetText(LPCTSTR s);
   //
   static DWORD CALLBACK StreamReadCallback(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG FAR *pcb);

// Variables
protected:
   BOOL  m_bFocus;
   HWND  m_hwndParent;
   HWND  m_hWnd;
   DWORD m_dwViewMode;
   DWORD m_dwBandID;
   CComQIPtr<IInputObjectSite, &IID_IInputObjectSite> m_spSite;
   CComQIPtr<IWebBrowser2> m_spWebBrowser2;
   CComQIPtr<IConnectionPointContainer> m_spCPC;
   //
   CTreeViewCtrl m_ctlTree;
   CRichEditCtrl m_ctlText;
   CImageList m_ctlImages;
   //
   CFindReplaceDialog* m_pFindDlg;
   CSimpleArray<tEntry> m_Entries;
   LPRICHEDITOLE m_pRichEditOle;
   CRtfBitmap m_bmp;
   DWORD m_dwCookie;
   //
   static CString m_sStream;
   static long    m_lStreamIndex;
};

#endif //__HTMLBAR_H_
